'use strict';

Object.defineProperty(exports, "__esModule", {
    value: true
});

var _typeof2 = require('babel-runtime/helpers/typeof');

var _typeof3 = _interopRequireDefault(_typeof2);

var _defineProperty2 = require('babel-runtime/helpers/defineProperty');

var _defineProperty3 = _interopRequireDefault(_defineProperty2);

var _assign = require('babel-runtime/core-js/object/assign');

var _assign2 = _interopRequireDefault(_assign);

var _toArray2 = require('babel-runtime/helpers/toArray');

var _toArray3 = _interopRequireDefault(_toArray2);

var _getIterator2 = require('babel-runtime/core-js/get-iterator');

var _getIterator3 = _interopRequireDefault(_getIterator2);

var _map = require('babel-runtime/core-js/map');

var _map2 = _interopRequireDefault(_map);

exports.node = node;
exports.demux = demux;

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var INITIAL_ACTION = '@@redux-combiner/INIT';

function node(initial) {

    checkInitial(initial, node);

    var reducersPaths = getReducersPaths(initial);

    var fn = function fn() {
        var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : getDefault(initial);
        var action = arguments[1];


        state = applyOwnReducers(state, action, fn.reducers);
        state = applyInnerReducers(state, action, reducersPaths);

        return state;
    };

    Object.defineProperty(fn, 'reducers', { value: new _map2.default() });
    Object.defineProperty(fn, 'on', { value: addActionsReducers });

    return fn;
}

function demux(initial, schema, options) {

    checkInitial(initial, demux);

    var reducersPaths = getReducersPaths(schema);

    var getChildKey = createActionKeyGetter(options);

    var fn = function fn() {
        var state = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : getDefault(initial);
        var action = arguments[1];


        state = applyOwnReducers(state, action, fn.reducers);
        state = applyChildInnerReducers(state, action, reducersPaths, getChildKey);

        return state;
    };

    Object.defineProperty(fn, 'reducers', { value: new _map2.default() });
    Object.defineProperty(fn, 'on', { value: addActionsReducers });

    return fn;
}

function checkInitial(initial, frame) {
    if (typeof initial === 'undefined') {
        var e = new Error('node must be initialized.');
        Error.captureStackTrace(e, frame);
        throw e;
    }
}

function getDefault(initial) {

    if (isFunction(initial)) {
        return initial(undefined, { type: INITIAL_ACTION });
    }

    if (isObject(initial)) {
        for (var k in initial) {
            if (initial.hasOwnProperty(k)) {
                initial[k] = getDefault(initial[k]);
            }
        }
    }

    return initial;
}

function getReducersPaths(reducer) {
    var reducersPaths = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var path = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : [];


    if (isFunction(reducer)) {
        reducersPaths.push({ path: path, reducer: reducer });
    }

    if (isObject(reducer)) {
        for (var k in reducer) {
            if (reducer.hasOwnProperty(k)) {
                getReducersPaths(reducer[k], reducersPaths, path.concat(k));
            }
        }
    }

    return reducersPaths;
}

function applyOwnReducers(state, action, actionReducers) {

    var reducers = actionReducers.get(action.type);

    if (reducers) {
        state = reducers.reduce(function (state, reducer) {
            return reducer(state, action);
        }, state);
    }

    return state;
}

function applyInnerReducers(state, action, reducersPaths) {
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
        outer: for (var _iterator = (0, _getIterator3.default)(reducersPaths), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var rp = _step.value;


            var pointer = state;

            var _iteratorNormalCompletion2 = true;
            var _didIteratorError2 = false;
            var _iteratorError2 = undefined;

            try {
                for (var _iterator2 = (0, _getIterator3.default)(rp.path), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
                    var k = _step2.value;


                    if (!isObject(pointer)) {
                        continue outer;
                    }
                    pointer = pointer[k];
                }
            } catch (err) {
                _didIteratorError2 = true;
                _iteratorError2 = err;
            } finally {
                try {
                    if (!_iteratorNormalCompletion2 && _iterator2.return) {
                        _iterator2.return();
                    }
                } finally {
                    if (_didIteratorError2) {
                        throw _iteratorError2;
                    }
                }
            }

            var result = rp.reducer(pointer, action);

            if (result !== pointer) {
                state = restate(state, rp.path, result);
            }
        }
    } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion && _iterator.return) {
                _iterator.return();
            }
        } finally {
            if (_didIteratorError) {
                throw _iteratorError;
            }
        }
    }

    return state;
}

function applyChildInnerReducers(state, action, reducersPaths, getChildKey) {

    var childKey = getChildKey(state, action);

    if (typeof childKey === 'undefined' || !state.hasOwnProperty(childKey)) {
        return state;
    }

    var child = state[childKey];

    var newChild = applyInnerReducers(child, action, reducersPaths);

    if (child !== newChild) {
        state = restate(state, [childKey], newChild);
    }

    return state;
}

function createActionKeyGetter() {
    var _ref = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {},
        actionKey = _ref.actionKey,
        getKey = _ref.getKey;

    if (isFunction(getKey)) {
        return getKey;
    }

    if (!actionKey) {
        return getDefaultActionKey;
    }

    if (!Array.isArray(actionKey)) {
        if (typeof actionKey === 'string') {
            actionKey = actionKey.split('.');
        } else {
            actionKey = [actionKey];
        }
    }

    return function (state, action) {
        return getValueByPath(action, actionKey);
    };
}

function getDefaultActionKey(state, action) {
    return Array.isArray(state) ? action.index : action.id;
}

function getValueByPath(obj, path) {

    var pointer = obj;

    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
        for (var _iterator3 = (0, _getIterator3.default)(path), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
            var p = _step3.value;


            if (!pointer.hasOwnProperty(p)) {
                return;
            }

            pointer = pointer[p];
        }
    } catch (err) {
        _didIteratorError3 = true;
        _iteratorError3 = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion3 && _iterator3.return) {
                _iterator3.return();
            }
        } finally {
            if (_didIteratorError3) {
                throw _iteratorError3;
            }
        }
    }

    return pointer;
}

function restate(state, path, value) {

    if (!path.length) {
        return value;
    }

    var result = Array.isArray(state) ? [] : {};

    var _path = (0, _toArray3.default)(path),
        key = _path[0],
        restPath = _path.slice(1);

    (0, _assign2.default)(result, state, (0, _defineProperty3.default)({}, key, restate(state[key], restPath, value)));

    return result;
}

function addActionsReducers(actions, reducers) {
    var _this = this;

    if (!Array.isArray(actions)) {
        actions = [actions];
    }

    reducers = getReducers(reducers);

    actions.forEach(function (actionType) {

        reducers.forEach(function (reducer) {

            var actionReducers = _this.reducers.get(actionType);

            if (!actionReducers) {
                actionReducers = [];
                _this.reducers.set(actionType, actionReducers);
            }

            actionReducers.push(reducer);
        });
    });

    return this;
}

function getReducers(reducers) {

    if (isFunction(reducers)) {
        return [reducers];
    }

    if (!Array.isArray(reducers)) {
        return [function () {
            return reducers;
        }];
    }

    var containsReducer = false;

    var _iteratorNormalCompletion4 = true;
    var _didIteratorError4 = false;
    var _iteratorError4 = undefined;

    try {
        for (var _iterator4 = (0, _getIterator3.default)(reducers), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
            var reducer = _step4.value;

            if (isFunction(reducer)) {
                containsReducer = true;
                break;
            }
        }
    } catch (err) {
        _didIteratorError4 = true;
        _iteratorError4 = err;
    } finally {
        try {
            if (!_iteratorNormalCompletion4 && _iterator4.return) {
                _iterator4.return();
            }
        } finally {
            if (_didIteratorError4) {
                throw _iteratorError4;
            }
        }
    }

    if (containsReducer) {
        return reducers.map(function (reducer) {
            return isFunction(reducer) ? reducer : function () {
                return reducer;
            };
        });
    }

    return [function () {
        return reducers;
    }];
}

function isObject(obj) {
    return obj && (typeof obj === 'undefined' ? 'undefined' : (0, _typeof3.default)(obj)) === 'object';
}

function isFunction(fn) {
    return typeof fn === 'function';
}